from __future__ import annotations

import asyncio
import contextlib
import logging
import multiprocessing as mp
import sys
import threading
from multiprocessing.context import BaseContext
from typing import Any, Callable, Coroutine

from .. import utils
from ..job import JobContext, JobProcess, RunningJobInfo
from ..log import logger
from . import channel, proc_main, proto


class LogQueueListener:
    _sentinel = None

    def __init__(
        self, queue: mp.Queue, prepare_fnc: Callable[[logging.LogRecord], None]
    ):
        self._thread: threading.Thread | None = None
        self._q = queue
        self._prepare_fnc = prepare_fnc

    def start(self) -> None:
        self._thread = t = threading.Thread(
            target=self._monitor, daemon=True, name="log_listener"
        )
        t.start()

    def stop(self) -> None:
        if self._thread is None:
            return
        self._q.put_nowait(self._sentinel)
        self._thread.join()
        self._thread = None

    def handle(self, record: logging.LogRecord) -> None:
        self._prepare_fnc(record)

        lger = logging.getLogger(record.name)
        if not lger.isEnabledFor(record.levelno):
            return

        lger.callHandlers(record)

    def _monitor(self):
        while True:
            record = self._q.get()
            if record is self._sentinel:
                break

            self.handle(record)


class SupervisedProc:
    def __init__(
        self,
        *,
        initialize_process_fnc: Callable[[JobProcess], Any],
        job_entrypoint_fnc: Callable[[JobContext], Coroutine],
        initialize_timeout: float,
        close_timeout: float,
        mp_ctx: BaseContext,
        loop: asyncio.AbstractEventLoop,
    ) -> None:
        self._loop = loop
        log_q = mp_ctx.Queue()
        log_q.cancel_join_thread()
        mp_pch, mp_cch = mp_ctx.Pipe(duplex=True)

        self._pch = channel.AsyncProcChannel(
            conn=mp_pch, loop=self._loop, messages=proto.IPC_MESSAGES
        )

        self._initialize_timeout = initialize_timeout
        self._close_timeout = close_timeout
        self._proc_args = proto.ProcStartArgs(
            initialize_process_fnc=initialize_process_fnc,
            job_entrypoint_fnc=job_entrypoint_fnc,
            log_q=log_q,
            mp_cch=mp_cch,
            asyncio_debug=loop.get_debug(),
        )

        self._proc = mp_ctx.Process(  # type: ignore
            target=proc_main.main, args=(self._proc_args,), name="job_proc"
        )
        self._running_job: RunningJobInfo | None = None

        self._exitcode: int | None = None
        self._pid: int | None = self._proc.pid

        self._main_atask: asyncio.Task[None] | None = None
        self._closing = False
        self._kill_sent = False
        self._initialize_fut = asyncio.Future[None]()

    @property
    def exitcode(self) -> int | None:
        return self._exitcode

    @property
    def killed(self) -> bool:
        return self._kill_sent

    @property
    def pid(self) -> int | None:
        return self._pid

    @property
    def started(self) -> bool:
        return self._main_atask is not None

    @property
    def start_arguments(self) -> Any | None:
        return self._proc_args.user_arguments

    @start_arguments.setter
    def start_arguments(self, value: Any | None) -> None:
        self._proc_args.user_arguments = value

    @property
    def running_job(self) -> RunningJobInfo | None:
        return self._running_job

    def start(self) -> None:
        """start the job process"""
        if self.started:
            raise RuntimeError("process already started")

        if self._closing:
            raise RuntimeError("process is closed")

        def _add_proc_ctx_log(record: logging.LogRecord) -> None:
            extra = self.logging_extra()
            for key, value in extra.items():
                setattr(record, key, value)

        log_listener = LogQueueListener(self._proc_args.log_q, _add_proc_ctx_log)
        log_listener.start()

        self._proc.start()
        self._pid = self._proc.pid
        self._join_fut = asyncio.Future[None]()

        def _sync_run():
            self._proc.join()
            log_listener.stop()
            self._loop.call_soon_threadsafe(self._join_fut.set_result, None)

        thread = threading.Thread(target=_sync_run)
        thread.start()
        self._main_atask = asyncio.create_task(self._main_task())

    async def join(self) -> None:
        """wait for the job process to finish"""
        if not self.started:
            raise RuntimeError("process not started")

        if self._main_atask:
            await asyncio.shield(self._main_atask)

    async def initialize(self) -> None:
        """initialize the job process, this is calling the user provided initialize_process_fnc
        raise asyncio.TimeoutError if initialization times out"""
        await self._pch.asend(proto.InitializeRequest())

        # wait for the process to become ready
        try:
            init_res = await asyncio.wait_for(
                self._pch.arecv(), timeout=self._initialize_timeout
            )
            assert isinstance(
                init_res, proto.InitializeResponse
            ), "first message must be InitializeResponse"
        except asyncio.TimeoutError:
            self._initialize_fut.set_exception(
                asyncio.TimeoutError("process initialization timed out")
            )
            logger.error(
                "initialization timed out, killing job", extra=self.logging_extra()
            )
            self._send_kill_signal()
            raise
        else:
            self._initialize_fut.set_result(None)

    async def aclose(self) -> None:
        """attempt to gracefully close the job process"""
        if not self.started:
            raise RuntimeError("process not started")

        self._closing = True
        with contextlib.suppress(channel.ChannelClosed):
            await self._pch.asend(proto.ShutdownRequest())

        try:
            if self._main_atask:
                await asyncio.wait_for(
                    asyncio.shield(self._main_atask), timeout=self._close_timeout
                )
        except asyncio.TimeoutError:
            logger.error(
                "process did not exit in time, killing job", extra=self.logging_extra()
            )
            self._send_kill_signal()

        if self._main_atask:
            await asyncio.shield(self._main_atask)

    async def kill(self) -> None:
        """forcefully kill the job process"""
        if not self.started:
            raise RuntimeError("process not started")

        self._closing = True
        self._send_kill_signal()
        if self._main_atask:
            await asyncio.shield(self._main_atask)

    async def launch_job(self, info: RunningJobInfo) -> None:
        """start/assign a job to the process"""
        if self._running_job is not None:
            raise RuntimeError("process already has a running job")

        self._running_job = info
        start_req = proto.StartJobRequest()
        start_req.running_job = info
        await self._pch.asend(start_req)

    def _send_kill_signal(self) -> None:
        """forcefully kill the job process"""
        try:
            if not self._proc.is_alive():
                return
        except ValueError:
            return

        logger.debug("killing job process", extra=self.logging_extra())
        if sys.platform == "win32":
            self._proc.terminate()
        else:
            self._proc.kill()

        self._kill_sent = True

    @utils.log_exceptions(logger=logger)
    async def _main_task(self) -> None:
        try:
            await self._initialize_fut
        except asyncio.TimeoutError:
            # this happens when the initialization takes longer than self._initialize_timeout
            pass

        # the process is killed if it doesn't respond to pings within this time
        pong_timeout = utils.aio.sleep(proto.PING_TIMEOUT)
        ping_task = asyncio.create_task(self._ping_pong_task(pong_timeout))
        monitor_task = asyncio.create_task(self._monitor_task(pong_timeout))

        await self._join_fut
        self._exitcode = self._proc.exitcode
        self._proc.close()
        await utils.aio.gracefully_cancel(ping_task, monitor_task)

        await self._pch.aclose()

        if self._exitcode != 0 and not self._kill_sent:
            logger.error(
                f"job process exited with non-zero exit code {self.exitcode}",
                extra=self.logging_extra(),
            )

    @utils.log_exceptions(logger=logger)
    async def _monitor_task(self, pong_timeout: utils.aio.Sleep) -> None:
        while True:
            msg = await self._pch.arecv()

            if isinstance(msg, proto.PongResponse):
                delay = utils.time_ms() - msg.timestamp
                if delay > proto.HIGH_PING_THRESHOLD * 1000:
                    logger.warning(
                        "job process is unresponsive",
                        extra={"delay": delay, **self.logging_extra()},
                    )

                with contextlib.suppress(utils.aio.SleepFinished):
                    pong_timeout.reset()

            if isinstance(msg, proto.Exiting):
                logger.debug(
                    "job exiting", extra={"reason": msg.reason, **self.logging_extra()}
                )

    @utils.log_exceptions(logger=logger)
    async def _ping_pong_task(self, pong_timeout: utils.aio.Sleep) -> None:
        ping_interval = utils.aio.interval(proto.PING_INTERVAL)

        async def _send_ping_co():
            while True:
                await ping_interval.tick()
                try:
                    await self._pch.asend(proto.PingRequest(timestamp=utils.time_ms()))
                except channel.ChannelClosed:
                    break

        async def _pong_timeout_co():
            await pong_timeout
            logger.error("job is unresponsive, killing job", extra=self.logging_extra())
            self._send_kill_signal()

        await asyncio.gather(_send_ping_co(), _pong_timeout_co())

    def logging_extra(self) -> dict:
        extra: dict = {
            "pid": self.pid,
        }
        if self._running_job:
            extra["job_id"] = self._running_job.job.id

        return extra



from __future__ import annotations

import asyncio
import contextlib
import time
from collections import deque
from dataclasses import dataclass
from typing import Awaitable, Callable, Deque, Optional, Union

from livekit import rtc

from .. import tokenize, utils
from ..log import logger
from . import _utils

# 3.83 is the "baseline", the number of hyphens per second TTS returns in avg.
STANDARD_SPEECH_RATE = 3.83


WillForwardTranscription = Callable[
    ["TTSSegmentsForwarder", rtc.Transcription],
    Union[rtc.Transcription, Awaitable[Optional[rtc.Transcription]]],
]


def _default_will_forward_transcription(
    fwd: TTSSegmentsForwarder, transcription: rtc.Transcription
) -> rtc.Transcription:
    return transcription


@dataclass
class _TTSOptions:
    room: rtc.Room
    participant_identity: str
    track_id: str
    language: str
    speed: float
    word_tokenizer: tokenize.WordTokenizer
    sentence_tokenizer: tokenize.SentenceTokenizer
    hyphenate_word: Callable[[str], list[str]]
    new_sentence_delay: float
    will_forward_transcription: WillForwardTranscription


@dataclass
class _SegmentData:
    segment_index: int
    sentence_stream: tokenize.SentenceStream
    pushed_text: str = ""
    pushed_duration: float = 0.0
    real_speed: float | None = None
    processed_sentences: int = 0
    processed_hyphens: int = 0
    validated: bool = False
    forward_start_time: float | None = 0.0


@dataclass
class _FormingSegments:
    audio: _SegmentData
    text: _SegmentData
    q: deque[_SegmentData]


class TTSSegmentsForwarder:
    """
    Forward TTS transcription to the users. This class tries to imitate the right timing of
    speech with the synthesized text. The first estimation is based on the speed argument. Once
    we have received the full audio of a specific text segment, we recalculate the avg speech
    speed using the length of the text & audio and catch up/ slow down the transcription if needed.
    """

    def __init__(
        self,
        *,
        room: rtc.Room,
        participant: rtc.Participant | str,
        track: rtc.Track | rtc.TrackPublication | str | None = None,
        language: str = "",
        speed: float = 1.0,
        new_sentence_delay: float = 0.4,
        word_tokenizer: tokenize.WordTokenizer = tokenize.basic.WordTokenizer(),
        sentence_tokenizer: tokenize.SentenceTokenizer = tokenize.basic.SentenceTokenizer(),
        hyphenate_word: Callable[[str], list[str]] = tokenize.basic.hyphenate_word,
        will_forward_transcription: WillForwardTranscription = _default_will_forward_transcription,
        loop: asyncio.AbstractEventLoop | None = None,
    ):
        """
        Args:
            room: room where the transcription will be sent
            participant: participant or identity that is pushing the TTS
            track: track where the TTS audio is being sent
            language: language of the text
            speed: average speech speed in characters per second (used by default if the full audio is not received yet)
            new_sentence_delay: delay in seconds between sentences
            auto_playout: if True, the forwarder will automatically start the transcription once the
                first audio frame is received. If False, you need to call segment_playout_started
                to start the transcription.
            word_tokenizer: word tokenizer used to split the text into words
            sentence_tokenizer: sentence tokenizer used to split the text into sentences
            hyphenate_word: function that returns a list of hyphens for a given word

        """
        identity = participant if isinstance(participant, str) else participant.identity

        if track is None:
            track = _utils.find_micro_track_id(room, identity)
        elif isinstance(track, (rtc.TrackPublication, rtc.Track)):
            track = track.sid

        speed = speed * STANDARD_SPEECH_RATE
        self._opts = _TTSOptions(
            room=room,
            participant_identity=identity,
            track_id=track,
            language=language,
            speed=speed,
            word_tokenizer=word_tokenizer,
            sentence_tokenizer=sentence_tokenizer,
            hyphenate_word=hyphenate_word,
            new_sentence_delay=new_sentence_delay,
            will_forward_transcription=will_forward_transcription,
        )
        self._closed = False
        self._loop = loop or asyncio.get_event_loop()
        self._close_future = asyncio.Future[None]()

        self._next_segment_index = 0
        self._playing_seg_index = -1
        self._finshed_seg_index = -1

        first_segment = self._create_segment()
        segments_q: Deque[_SegmentData] = deque()
        segments_q.append(first_segment)

        self._forming_segments = _FormingSegments(
            audio=first_segment, text=first_segment, q=segments_q
        )

        self._seg_queue = asyncio.Queue[Optional[_SegmentData]]()
        self._seg_queue.put_nowait(first_segment)
        self._main_atask = self._loop.create_task(self._main_task())
        self._task_set = utils.aio.TaskSet(loop)

        self._played_text = ""

    def segment_playout_started(self) -> None:
        """
        Notify that the playout of the audio segment has started.
        This will start forwarding the transcription for the current segment.
        """
        self._check_not_closed()
        self._playing_seg_index += 1

    def segment_playout_finished(self) -> None:
        """
        Notify that the playout of the audio segment has finished.
        This will catchup and directly send the final transcription in case the forwarder is too
        late.
        """
        self._check_not_closed()
        self._finshed_seg_index += 1

    def push_audio(self, frame: rtc.AudioFrame) -> None:
        self._check_not_closed()
        frame_duration = frame.samples_per_channel / frame.sample_rate
        cur_seg = self._forming_segments.audio
        cur_seg.pushed_duration += frame_duration
        cur_seg.validated = True

    def mark_audio_segment_end(self) -> None:
        self._check_not_closed()
        try:
            # get last ended segment (text always end before audio)
            seg = self._forming_segments.q.popleft()
        except IndexError:
            raise IndexError(
                "mark_audio_segment_end called before any mark_text_segment_end"
            )

        if seg.pushed_duration > 0.0:
            seg.real_speed = (
                len(self._calc_hyphens(seg.pushed_text)) / seg.pushed_duration
            )

        seg.validated = True
        self._forming_segments.audio = self._forming_segments.q[0]

    def push_text(self, text: str) -> None:
        self._check_not_closed()
        cur_seg = self._forming_segments.text
        cur_seg.pushed_text += text
        cur_seg.sentence_stream.push_text(text)

    def mark_text_segment_end(self) -> None:
        self._check_not_closed()
        stream = self._forming_segments.text.sentence_stream
        stream.end_input()

        # create a new segment on "mark_text_segment_end"
        # further text can already be pushed even if mark_audio_segment_end has not been
        # called yet
        new_seg = self._create_segment()
        self._forming_segments.text = new_seg
        self._forming_segments.q.append(new_seg)
        self._seg_queue.put_nowait(new_seg)

    @property
    def closed(self) -> bool:
        return self._closed

    @property
    def played_text(self) -> str:
        return self._played_text

    async def aclose(self) -> None:
        if self._closed:
            return

        self._closed = True
        self._close_future.set_result(None)
        self._seg_queue.put_nowait(None)

        for seg in self._forming_segments.q:
            await seg.sentence_stream.aclose()

        await self._task_set.aclose()
        await self._main_atask

    @utils.log_exceptions(logger=logger)
    async def _main_task(self) -> None:
        """Main task that forwards the transcription to the room."""
        rtc_seg_q = asyncio.Queue[Optional[rtc.TranscriptionSegment]]()

        @utils.log_exceptions(logger=logger)
        async def _forward_task():
            while True:
                seg = await rtc_seg_q.get()
                if seg is None:
                    break

                base_transcription = rtc.Transcription(
                    participant_identity=self._opts.participant_identity,
                    track_sid=self._opts.track_id,
                    segments=[seg],  # no history for now
                )

                transcription = self._opts.will_forward_transcription(
                    self, base_transcription
                )
                if asyncio.iscoroutine(transcription):
                    transcription = await transcription

                # fallback to default impl if no custom/user stream is returned
                if not isinstance(transcription, rtc.Transcription):
                    transcription = _default_will_forward_transcription(
                        self, base_transcription
                    )

                if transcription.segments and self._opts.room.isconnected():
                    await self._opts.room.local_participant.publish_transcription(
                        transcription
                    )

        forward_task = asyncio.create_task(_forward_task())

        while True:
            seg = await self._seg_queue.get()
            if seg is None:
                break

            # wait until the segment is validated and has started playing
            while not self._closed:
                if seg.validated and self._playing_seg_index >= seg.segment_index:
                    break

                await self._sleep_if_not_closed(0.1)

            sentence_stream = seg.sentence_stream
            seg.forward_start_time = time.time()

            async for ev in sentence_stream:
                await self._sync_sentence_co(seg, ev.token, rtc_seg_q)

        rtc_seg_q.put_nowait(None)
        await forward_task

    async def _sync_sentence_co(
        self,
        seg: _SegmentData,
        tokenized_sentence: str,
        rtc_seg_q: asyncio.Queue[Optional[rtc.TranscriptionSegment]],
    ):
        """Synchronize the transcription with the audio playout for a given sentence."""
        assert seg.forward_start_time is not None

        # put each sentence in a different transcription segment
        seg_id = _utils.segment_uuid()
        words = self._opts.word_tokenizer.tokenize(text=tokenized_sentence)
        processed_words: list[str] = []

        og_text = self._played_text
        for word in words:
            if seg.segment_index <= self._finshed_seg_index:
                # playout of the audio segment already finished
                # break the loop and send the final transcription
                break

            if self._closed:
                # transcription closed, early
                return

            word_hyphens = len(self._opts.hyphenate_word(word))
            processed_words.append(word)

            # elapsed time since the start of the seg
            elapsed_time = time.time() - seg.forward_start_time
            text = self._opts.word_tokenizer.format_words(processed_words)

            speed = self._opts.speed
            if seg.real_speed is not None:
                speed = seg.real_speed
                estimated_pauses_s = (
                    seg.processed_sentences * self._opts.new_sentence_delay
                )
                hyph_pauses = estimated_pauses_s * speed

                target_hyphens = round(speed * elapsed_time)
                dt = target_hyphens - seg.processed_hyphens - hyph_pauses
                to_wait_hyphens = max(0.0, word_hyphens - dt)
                delay = to_wait_hyphens / speed
            else:
                delay = word_hyphens / speed

            first_delay = min(delay / 2, 2 / speed)
            await self._sleep_if_not_closed(first_delay)
            rtc_seg_q.put_nowait(
                rtc.TranscriptionSegment(
                    id=seg_id,
                    text=text,
                    start_time=0,
                    end_time=0,
                    final=False,
                    language=self._opts.language,
                )
            )
            self._played_text = f"{og_text} {text}"
            await self._sleep_if_not_closed(delay - first_delay)
            seg.processed_hyphens += word_hyphens

        rtc_seg_q.put_nowait(
            rtc.TranscriptionSegment(
                id=seg_id,
                text=tokenized_sentence,
                start_time=0,
                end_time=0,
                final=True,
                language=self._opts.language,
            )
        )
        self._played_text = f"{og_text} {tokenized_sentence}"

        await self._sleep_if_not_closed(self._opts.new_sentence_delay)
        seg.processed_sentences += 1

    async def _sleep_if_not_closed(self, delay: float) -> None:
        with contextlib.suppress(asyncio.TimeoutError):
            await asyncio.wait([self._close_future], timeout=delay)

    def _calc_hyphens(self, text: str) -> list[str]:
        hyphens: list[str] = []
        words = self._opts.word_tokenizer.tokenize(text=text)
        for word in words:
            new = self._opts.hyphenate_word(word)
            hyphens.extend(new)

        return hyphens

    def _create_segment(self) -> _SegmentData:
        data = _SegmentData(
            segment_index=self._next_segment_index,
            sentence_stream=self._opts.sentence_tokenizer.stream(),
        )
        self._next_segment_index += 1
        return data

    def _check_not_closed(self) -> None:
        if self._closed:
            raise RuntimeError("TTSForwarder is closed")



import time
import uuid
from typing import List, Union

from livekit import rtc

AudioBuffer = Union[List[rtc.AudioFrame], rtc.AudioFrame]


def merge_frames(buffer: AudioBuffer) -> rtc.AudioFrame:
    """
    Merges one or more AudioFrames into a single one
    Args:
        buffer: either a rtc.AudioFrame or a list of rtc.AudioFrame
    """
    if isinstance(buffer, list):
        # merge all frames into one
        if len(buffer) == 0:
            raise ValueError("buffer is empty")

        sample_rate = buffer[0].sample_rate
        num_channels = buffer[0].num_channels
        samples_per_channel = 0
        data = b""
        for frame in buffer:
            if frame.sample_rate != sample_rate:
                raise ValueError("sample rate mismatch")

            if frame.num_channels != num_channels:
                raise ValueError("channel count mismatch")

            data += frame.data
            samples_per_channel += frame.samples_per_channel

        return rtc.AudioFrame(
            data=data,
            sample_rate=sample_rate,
            num_channels=num_channels,
            samples_per_channel=samples_per_channel,
        )

    return buffer


def time_ms() -> int:
    return int(time.time() * 1000)


def shortuuid() -> str:
    return str(uuid.uuid4().hex)[:12]



from __future__ import annotations

import asyncio
import time
from typing import Any, AsyncIterable, Callable, Union

from livekit import rtc

from .. import llm, tokenize, utils
from .. import transcription as agent_transcription
from .. import tts as text_to_speech
from .agent_playout import AgentPlayout, PlayoutHandle
from .log import logger

SpeechSource = Union[AsyncIterable[str], str]


class SynthesisHandle:
    def __init__(
        self,
        *,
        speech_id: str,
        speech_source: SpeechSource,
        agent_playout: AgentPlayout,
        tts: text_to_speech.TTS,
        transcription_fwd: agent_transcription.TTSSegmentsForwarder,
    ) -> None:
        self._speech_source, self._agent_playout, self._tts, self._tr_fwd = (
            speech_source,
            agent_playout,
            tts,
            transcription_fwd,
        )
        self._buf_ch = utils.aio.Chan[rtc.AudioFrame]()
        self._play_handle: PlayoutHandle | None = None
        self._interrupt_fut = asyncio.Future[None]()
        self._speech_id = speech_id

    @property
    def speech_id(self) -> str:
        return self._speech_id

    @property
    def tts_forwarder(self) -> agent_transcription.TTSSegmentsForwarder:
        return self._tr_fwd

    @property
    def validated(self) -> bool:
        return self._play_handle is not None

    @property
    def interrupted(self) -> bool:
        return self._interrupt_fut.done()

    @property
    def play_handle(self) -> PlayoutHandle | None:
        return self._play_handle

    def play(self) -> PlayoutHandle:
        """Validate the speech for playout"""
        if self.interrupted:
            raise RuntimeError("synthesis was interrupted")

        self._play_handle = self._agent_playout.play(
            self._speech_id, self._buf_ch, transcription_fwd=self._tr_fwd
        )
        return self._play_handle

    def interrupt(self) -> None:
        """Interrupt the speech"""
        if self.interrupted:
            return

        logger.debug(
            "interrupting synthesis/playout",
            extra={"speech_id": self.speech_id},
        )

        if self._play_handle is not None:
            self._play_handle.interrupt()

        self._interrupt_fut.set_result(None)


class AgentOutput:
    def __init__(
        self,
        *,
        room: rtc.Room,
        agent_playout: AgentPlayout,
        llm: llm.LLM,
        tts: text_to_speech.TTS,
    ) -> None:
        self._room, self._agent_playout, self._llm, self._tts = (
            room,
            agent_playout,
            llm,
            tts,
        )
        self._tasks = set[asyncio.Task[Any]]()

    @property
    def playout(self) -> AgentPlayout:
        return self._agent_playout

    async def aclose(self) -> None:
        for task in self._tasks:
            task.cancel()

        await asyncio.gather(*self._tasks, return_exceptions=True)

    def synthesize(
        self,
        *,
        speech_id: str,
        transcript: SpeechSource,
        transcription: bool,
        transcription_speed: float,
        sentence_tokenizer: tokenize.SentenceTokenizer,
        word_tokenizer: tokenize.WordTokenizer,
        hyphenate_word: Callable[[str], list[str]],
    ) -> SynthesisHandle:
        def _will_forward_transcription(
            fwd: agent_transcription.TTSSegmentsForwarder,
            transcription: rtc.Transcription,
        ):
            if not transcription:
                transcription.segments = []

            return transcription

        transcription_fwd = agent_transcription.TTSSegmentsForwarder(
            room=self._room,
            participant=self._room.local_participant,
            speed=transcription_speed,
            sentence_tokenizer=sentence_tokenizer,
            word_tokenizer=word_tokenizer,
            hyphenate_word=hyphenate_word,
            will_forward_transcription=_will_forward_transcription,
        )

        handle = SynthesisHandle(
            speech_source=transcript,
            agent_playout=self._agent_playout,
            tts=self._tts,
            transcription_fwd=transcription_fwd,
            speech_id=speech_id,
        )

        task = asyncio.create_task(self._synthesize_task(handle))
        self._tasks.add(task)
        task.add_done_callback(self._tasks.remove)
        return handle

    @utils.log_exceptions(logger=logger)
    async def _synthesize_task(self, handle: SynthesisHandle) -> None:
        """Synthesize speech from the source"""
        if isinstance(handle._speech_source, str):
            co = _str_synthesis_task(handle._speech_source, handle)
        else:
            co = _stream_synthesis_task(handle._speech_source, handle)

        synth = asyncio.create_task(co)
        synth.add_done_callback(lambda _: handle._buf_ch.close())
        try:
            _ = await asyncio.wait(
                [synth, handle._interrupt_fut], return_when=asyncio.FIRST_COMPLETED
            )
        finally:
            await utils.aio.gracefully_cancel(synth)


@utils.log_exceptions(logger=logger)
async def _str_synthesis_task(text: str, handle: SynthesisHandle) -> None:
    """synthesize speech from a string"""
    if not handle.tts_forwarder.closed:
        handle.tts_forwarder.push_text(text)
        handle.tts_forwarder.mark_text_segment_end()

    start_time = time.time()
    first_frame = True

    try:
        async for audio in handle._tts.synthesize(text):
            if first_frame:
                first_frame = False
                logger.debug(
                    "received first TTS frame",
                    extra={
                        "speech_id": handle.speech_id,
                        "elapsed": round(time.time() - start_time, 3),
                        "streamed": False,
                    },
                )

            frame = audio.frame

            handle._buf_ch.send_nowait(frame)
            if not handle.tts_forwarder.closed:
                handle.tts_forwarder.push_audio(frame)

    finally:
        if not handle.tts_forwarder.closed:
            handle.tts_forwarder.mark_audio_segment_end()


@utils.log_exceptions(logger=logger)
async def _stream_synthesis_task(
    streamed_text: AsyncIterable[str], handle: SynthesisHandle
) -> None:
    """synthesize speech from streamed text"""

    @utils.log_exceptions(logger=logger)
    async def _read_generated_audio_task():
        start_time = time.time()
        first_frame = True
        async for audio in tts_stream:
            if first_frame:
                first_frame = False
                logger.debug(
                    "first TTS frame",
                    extra={
                        "speech_id": handle.speech_id,
                        "elapsed": round(time.time() - start_time, 3),
                        "streamed": True,
                    },
                )

            if not handle._tr_fwd.closed:
                handle._tr_fwd.push_audio(audio.frame)

            handle._buf_ch.send_nowait(audio.frame)

        if handle._tr_fwd and not handle._tr_fwd.closed:
            # mark_audio_segment_end must be called *after* mart_text_segment_end
            handle._tr_fwd.mark_audio_segment_end()

    # otherwise, stream the text to the TTS
    tts_stream = handle._tts.stream()
    read_atask: asyncio.Task | None = None

    try:
        async for seg in streamed_text:
            if not handle.tts_forwarder.closed:
                handle.tts_forwarder.push_text(seg)

            if read_atask is None:
                # start the task when we receive the first text segment (so start_time is more accurate)
                read_atask = asyncio.create_task(_read_generated_audio_task())

            tts_stream.push_text(seg)

        if not handle.tts_forwarder.closed:
            handle.tts_forwarder.mark_text_segment_end()

        tts_stream.end_input()

        if read_atask is not None:
            await read_atask
    finally:
        if read_atask is not None:
            await utils.aio.gracefully_cancel(read_atask)

        await tts_stream.aclose()



def split_paragraphs(text: str) -> list[str]:
    sep = "\n\n"

    paragraphs = text.split(sep)
    new_paragraphs = []
    for p in paragraphs:
        p = p.strip()
        if not p:
            continue
        new_paragraphs.append(p)

    return new_paragraphs
